/* nwall -- chat-friendly wall program.
 * Copyright (C) 1999 Nick Moffitt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <utmp.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <readline/readline.h>
#include <readline/history.h>

#define EVER ;;
#define IGNOREUSER "sleeper"
#define LOGFILE "/var/log/nwall.log"
#define OPTOUTFILE "/var/lock/TTYXXNN"

/* forward declarations */
void NwallEnter(char *, char *);
void NwallExit(char *, char *);
struct iovec NwallFormatMessage(struct iovec, char *, char *);
void NwallWriteAll(struct iovec);
void SignalHandler(int);

char *ttymsg();

/* global variables (I know, I know...) */
char *progname;

int main(int argc, char **argv)
{
	char *username, *ttystring, *ttyfilename;
	struct iovec message;

	username = getlogin();
	ttyfilename = ttyname(2);
	ttystring = ttyfilename + 5*sizeof(char);

	progname = argv[0];

	message.iov_base = (char *)NULL;
	message.iov_len = 0;


	/* Don't use tab-completion. */
	rl_bind_key('\t', rl_insert);

	/* Say hello to all the nice users */
	NwallEnter(username, ttystring);

	signal(SIGINT, SignalHandler);


	for(EVER) {
		/* If the buffer has already been allocated, return the memory
		 * to the free pool. 
	 	 */
		if (message.iov_base) {
			free(message.iov_base);
			message.iov_base = (char *)NULL;
		}
		/* Get a line from the user. */
		message.iov_base = (char *)readline ("nwall> ");

		/* If readline returned a null pointer (IE, ^D was pressed),
		 * exit with a pleasant message.
		 */
		if(!message.iov_base) {
			NwallExit(username, ttystring);
			exit(0); /* leave program */
		}

		/* since the pointer could have been null, we need to run
		 * strlen() after we check for that.
		 */
		message.iov_len = strlen((char *)message.iov_base) 
				+ sizeof(char);

		/* If the line has any text in it, save it on the history. */
		if ((char *)message.iov_base && *(char *)message.iov_base) {
			add_history ((char *)message.iov_base);
		}

		/* Add attribution (eg "nick@ttyp1> hello!").
		 * Eventually we'll format for 80 columns using GNU fmt's
		 * functions.
		 */
		message = NwallFormatMessage(message, username, ttystring);

		/* Go through the utmp and write the message to everyone
		 * logged in.  Also does logging and opt-outs.
		 */
		NwallWriteAll(message);
	}

} /* end of main() */

/*******************************************************************/


/* probably better to do these with a third string parameter and a
 * strncat somewhere, but I'm sleepy.
 */

/* print greeting message as user begins to wall */
void NwallEnter(char *username, char *ttystring)
{
		struct iovec hello;

		hello.iov_len = strlen(username) + strlen(ttystring) + 
			sizeof(char) + 
			strlen(" has joined the nwall conversation.\a\r\n") + 
			sizeof(char);
		hello.iov_base = (char *)malloc(hello.iov_len);

		snprintf((char *)hello.iov_base, hello.iov_len, 
			 "%s@%s has joined the nwall conversation.\a\r\n", 
			 username, ttystring);

		NwallWriteAll(hello);
				
		free(hello.iov_base);
		return;
}

/* show that user has left the nwall program */
void NwallExit(char *username, char *ttystring)
{
	struct iovec eof;

	eof.iov_len = strlen(username) + sizeof(char) + strlen(ttystring) 
		+ strlen(" has left the building.\a\r\n") + 
		sizeof(char);
	eof.iov_base = (char *)malloc(eof.iov_len);

	snprintf((char *)eof.iov_base, eof.iov_len, 
		 "%s@%s has left the building.\a\r\n", username, 
		 ttystring);

	NwallWriteAll(eof);

	free(eof.iov_base);
	return;
}

/* pretty-print the user's line of text, complete with allsortsfulla
 * info on username and tty (soon to include hostname if remote!)
 * Ultimately we'll use the functions from GNU fmt to format this to
 * 80 columns or something.
 */
struct iovec NwallFormatMessage(struct iovec inmsg, char *username, 
								char *ttystring)
{
		struct iovec outmsg;

		outmsg.iov_len = sizeof(char) + strlen(username) +
				sizeof(char) + strlen(ttystring) + 2*sizeof(char) +
				inmsg.iov_len + 2*sizeof(char);
		outmsg.iov_base = (char *)malloc(outmsg.iov_len);

		snprintf((char *)outmsg.iov_base, outmsg.iov_len, 
				 " %s@%s> %s\r\n", username, ttystring, inmsg.iov_base);

		free(inmsg.iov_base);
		return outmsg;
}

/* write to all ttys (searching through utmp and excluding opt-outs
 * and the mysterious "sleeper") and do logging.
 */
void NwallWriteAll(struct iovec msg)
{
	struct utmp *utmpptr;
	struct stat optout;
	size_t optoutlen;
	char *ttymsgerror, *optoutfile;
	int fd, writeverr;


	/* log the message right off the bat */
	fd = open(LOGFILE, O_WRONLY|O_APPEND|O_NONBLOCK);
	if((writeverr = writev(fd, &msg, 1)) == -1) {
		perror("nwall had a problem opening the log file");
	}
	close(fd);


	/* send one message to each logged-in user that hasn't opted out */
    setutent();

	/* For each entry in /var/utmp (each logged in user) */
    while((utmpptr = getutent())) {
 		optoutlen = strlen("/var/lock") + strlen(utmpptr->ut_line) + sizeof(char);
		optoutfile = (char *)malloc(optoutlen);
		snprintf(optoutfile, optoutlen, "/var/lock/%s", utmpptr->ut_line);

	/* Here come all the reasons not to send the message to a
	 * particular user:
	 */

		/* opted out? (test is whether optout file is owned by current
		 * user for that tty)
		 */
		if(!stat(optoutfile, &optout)) {
			if(0) {/* XXX usernames match XXX */
				continue;
			}
		}
		free(optoutfile);

		/* valid utmp entry?  not the mysterious "sleeper" user? */
        if (!utmpptr->ut_name[0] ||
            !strncmp(utmpptr->ut_name, IGNOREUSER, sizeof(utmpptr->ut_name)))
            continue;

		/* is this a proper user process? */
        if (utmpptr->ut_type != USER_PROCESS)
            continue;

		/* Actually send the message! */
        if ((ttymsgerror = ttymsg(&msg, 1, utmpptr->ut_line, 60*5)) != NULL)
		perror("nwall had a problem sending your message to a user");
    }

	/* close up our use of /var/utmp */
    endutent();

	return;
} /* end of NwallWriteAll */


void SignalHandler(int signal)
{
	if(signal == SIGINT) {
		NwallExit(getlogin(), ttyname(2) + 5*sizeof(char));
	}
	exit(15);
}




